/*
 * Copyright 1992 John L. Cwikla
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appears in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of John L. Cwikla or
 * University of Illinois not be used in advertising or publicity
 * pertaining to distribution of the software without specific, written
 * prior permission.  John L. Cwikla and University of Illinois make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 *
 * John L. Cwikla and University of Illinois disclaim all warranties with
 * regard to this software, including all implied warranties of
 * merchantability and fitness, in no event shall John L. Cwikla or
 * University of Illinois be liable for any special, indirect or
 * consequential damages or any damages whatsoever resulting from loss of
 * use, data or profits, whether in an action of contract, negligence or
 * other tortious action, arising out of or in connection with the use or
 * performance of this software.
 *
 * Author:
 * 	John L. Cwikla
 * 	Materials Research Laboratory Center for Computation
 * 	University Of Illinois at Urbana-Champaign
 *	104 S. Goodwin
 * 	Urbana, IL 61801
 * 
 * 	cwikla@uimrl7.mrl.uiuc.edu
 */

/*
** Image Widget
*/

#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <math.h>
#include <varargs.h>

#include "ImageP.h"
#include "Image.h"

/* For actions */
#if NeedFunctionProtoTypes
void Select(ImageWidget _w, XEvent *_event, String *__argv, int _argc); 
#else
void Select();
#endif

/* For widget internals */
#if NeedFunctionProtoTypes
static void Initialize(ImageWidget _request, ImageWidget _new);
static void Resize(ImageWidget _w);
static void Destroy(ImageWidget _w);
static void Redisplay(ImageWidget _w, XEvent *_event, Region _region);
static Boolean SetValues(ImageWidget _current, ImageWidget _request, ImageWidget _new);
static void MyXtWarning();
static void ClassInitialize();
#else
static void Initialize();
static void Resize();
static void Destroy();
static void Redisplay();
static Boolean SetValues();
static void MyXtWarning();
static void ClassInitialize();
#endif

/* For Misc */
#if NeedFunctionProtoTypes
static void CreateTheImage(ImageWidget _w);
static void CheckPosition(ImageWidget _w);
static Boolean CvtStringToGravity(Display *_display, XrmValuePtr _args, Cardinal *_numArgs, 
  XrmValuePtr _from, XrmValuePtr _to, XtPointer *_data);
#else
static void CreateTheImage();
static void CheckPosition();
static Boolean CvtStringToGravity();
#endif

static char ImageTranslations[] =
"<Btn1Down>:    select()\n\
 <Btn1Motion>:  select()";

static XtActionsRec ImageActions[] = 
{
  { "select", Select}
};

/* Just for fun and tests. */

#define NoDataWidth 40 
#define NoDataHeight 40 
static char NoDataPic[] = {
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
};

#define TheOffset(field) XtOffset(ImageWidget, image.field)

static XtResource ImageResources[] = 
{
  {XtNselectCallback, XtCCallback, XtRCallback, sizeof(caddr_t),
    TheOffset(selectCallback), XtRCallback, (caddr_t)NULL},
  {XtNdataWidth, XtCDataWidth, XtRInt, sizeof(int),
    TheOffset(width), XtRImmediate, (caddr_t)NoDataHeight},
  {XtNdataHeight, XtCDataHeight, XtRInt, sizeof(int),
    TheOffset(height), XtRImmediate, (caddr_t)NoDataWidth},
  {XtNdata, XtCData, XtRPointer, sizeof(caddr_t),
    TheOffset(data), XtRImmediate, (caddr_t)NoDataPic},
  {XtNgravity, XtCGravity, XtRGravity, sizeof(int),
    TheOffset(gravity), XtRImmediate, (caddr_t)CenterGravity}
};

#undef TheOffset

ImageClassRec imageClassRec = 
{
  {        /* CoreClassPart */
    (WidgetClass)&widgetClassRec,  /* superclass */
    "Image",                       /* class_name */
    sizeof(ImageRec),              /* widget_size */
    ClassInitialize,               /* class_initialize */
    NULL,                          /* class_part_initialize */
    FALSE,                         /* class_init */
    Initialize,                    /* initialize */
    NULL,                          /* initialize_hook */
    XtInheritRealize,              /* realize */
    ImageActions,                  /* actions */
    XtNumber(ImageActions),        /* num_actions */
    ImageResources,                /* resources */
    XtNumber(ImageResources),      /* num_resources */
    NULLQUARK,                     /* xrm_class */
    TRUE,                          /* compress_motion */
    TRUE,                          /* compress_exposure */
    TRUE,                          /* compress_enterleave */
    TRUE,                          /* visible_intress */
    Destroy,                       /* destroy */
    Resize,                        /* resize */
    Redisplay,                     /* expose */
    SetValues,                     /* set_values */
    NULL,                          /* set_values_hook */
    XtInheritSetValuesAlmost,      /* set_values_almost */
    NULL,                          /* get_values_hook */
    NULL,                          /* accept_focus */
    XtVersion,                     /* version */
    NULL,                          /* callback_private */
    ImageTranslations,             /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  { 
    0 /* empty */
  }
};

WidgetClass imageWidgetClass = (WidgetClass) &imageClassRec;

static void ClassInitialize()
{
  XtSetTypeConverter(XtRString, XtRGravity, CvtStringToGravity,
    (XtConvertArgList)NULL, 0, XtCacheAll, (XtDestructor)NULL);
}

#define IMAGE _new->image
#define CORE _new->core

static void Initialize(_request, _new)
ImageWidget _request;
ImageWidget _new;
{
  Display *display;

  display = XtDisplay(_new);

  if (IMAGE.visual == (Visual *)NULL)
    IMAGE.visual = DefaultVisual(display, DefaultScreen(display));

  if (_request->core.width == 0)
    CORE.width = IMAGE.width;
  if (_request->core.height == 0)
    CORE.height = IMAGE.height;

/* Read only gc */

  IMAGE.gc = XtGetGC(_new, NULL, 0);

  CreateTheImage(_new);
  CheckPosition(_new); 
}

#undef IMAGE 
#undef CORE

#define IMAGE _w->image

#define ABS(a) (((a) >= 0) ? (a) : (-1*(a)))
#define CENTERX(w) (w->core.width/2 - w->image.width/2)
#define CENTERY(w) (w->core.height/2 - w->image.height/2)
#define RIGHTJUSTIFY(w) (w->core.width - w->image.width)
#define BOTTOMJUSTIFY(w) (w->core.height - w->image.height)
#define DVISIBLE(a) MAX((a), 0)
#define SVISIBLE(a) (((a) < 0) ? ABS(a) : 0)
#define SVISWIDTH(w) MIN(w->core.width, w->image.width)
#define SVISHEIGHT(w) MIN(w->core.height, w->image.height)

static void CheckPosition(_w)
ImageWidget _w;
{
  switch (_w->image.gravity)
  {
    case NorthWestGravity:
         IMAGE.srcX = IMAGE.destX = 
         IMAGE.srcY = IMAGE.destY = 0;
         break;
    case NorthGravity:
         IMAGE.srcX = IMAGE.destX = CENTERX(_w);
         IMAGE.srcY = IMAGE.destY = 0;
         break;
    case NorthEastGravity:
         IMAGE.srcX = IMAGE.destX = RIGHTJUSTIFY(_w);
         IMAGE.srcY = IMAGE.destY = 0;
         break;
    case WestGravity:
	 IMAGE.srcX = IMAGE.destX = 0;
         IMAGE.srcY = IMAGE.destY = CENTERY(_w);
         break;
    case CenterGravity:
         IMAGE.srcX = IMAGE.destX = CENTERX(_w);
         IMAGE.srcY = IMAGE.destY = CENTERY(_w);
         break;
    case EastGravity:
         IMAGE.srcX = IMAGE.destX = RIGHTJUSTIFY(_w);
         IMAGE.srcY = IMAGE.destY = CENTERY(_w);
         break;
    case SouthWestGravity:
         IMAGE.srcX = IMAGE.destX = 0;
         IMAGE.srcY = IMAGE.destY = BOTTOMJUSTIFY(_w);
         break;
    case SouthGravity:
         IMAGE.srcX = IMAGE.destX = CENTERX(_w);
         IMAGE.srcY = IMAGE.destY = BOTTOMJUSTIFY(_w);
         break;
    case SouthEastGravity:
         IMAGE.srcX = IMAGE.destX = RIGHTJUSTIFY(_w);
         IMAGE.srcY = IMAGE.destY = BOTTOMJUSTIFY(_w);
         break;
    case StaticGravity: /* Don't do anything */
         return;
  }

  IMAGE.srcX = SVISIBLE(IMAGE.srcX);
  IMAGE.srcY = SVISIBLE(IMAGE.srcY);
  IMAGE.destX = DVISIBLE(IMAGE.destX);
  IMAGE.destY = DVISIBLE(IMAGE.destY);
}

static void Resize(_w)
ImageWidget _w;
{
  CheckPosition(_w);
}

static void Destroy(_w)
ImageWidget _w;
{
  XtFree(_w->image.visual);
  XDestroyImage(_w->image.ximage);
  XtReleaseGC((Widget)_w, _w->image.gc);
  XtRemoveAllCallbacks((Widget)_w, XtNselectCallback);
}

/* To come: merging the clipbox rect with the image at hand to see what needs to be
** redisplayed.
*/

static void Redisplay(_w, _event, _region)
ImageWidget _w; 
XEvent *_event; 
Region _region;
{
  XRectangle rect;

  XClipBox(_region, &rect);
  if (_w->core.visible)
    XPutImage(XtDisplay(_w), XtWindow(_w), _w->image.gc, _w->image.ximage,
      IMAGE.srcX, IMAGE.srcY, IMAGE.destX, IMAGE.destY,
      SVISWIDTH(_w), SVISHEIGHT(_w));
  XFlush(XtDisplay(_w));
}

static Boolean SetValues(_current, _request, _new)
ImageWidget _current; 
ImageWidget _request; 
ImageWidget _new;
{
  *_new = *_current; /* For now don't allow any changes */
#if 0
  Boolean redisplay = FALSE;
  Display *display = XtDisplay(_new);

  printf("SetValues\n");
  return redisplay;
#endif
  return FALSE;
}
#undef IMAGE 

/* This will screw up right now since I don't check to see if the
** coords are actually in the image yet.
*/

static void Select(_w, _event, _argv, _argc)
ImageWidget _w;
XEvent *_event;
String *_argv;
int *_argc;
{
  int x, y;
  ImageCallbackStruct ics;

  switch (_event->type)
  {
    case ButtonPress:
    case ButtonRelease:
      x = ((XButtonEvent *)_event)->x;
      y = ((XButtonEvent *)_event)->y;
      break;
    case KeyPress:
    case KeyRelease:
      x = ((XKeyEvent *)_event)->x;
      y = ((XKeyEvent *)_event)->y;
      break;
    case MotionNotify:
      x = ((XMotionEvent *)_event)->x;
      y = ((XMotionEvent *)_event)->y;
      break;
    default: return; break;
  }

  ics.reason = IMAGE_SELECT;
  ics.event = _event;
  ics.pixel = XGetPixel(_w->image.ximage, x, y);
  printf("%d ics.pixel\n", ics.pixel);
  XtCallCallbacks((Widget)_w, XtNselectCallback, &ics);
}

#define MAXSTRING 300

static void MyXtWarning(_format, va_alist)
char *_format; 
va_dcl   /* stupid define already has a ; on it */
{
  va_list parms;
  char dest[MAXSTRING];

  va_start(parms);
  vsprintf(dest, _format, parms);
  va_end(parms);

  XtWarning(dest);
}

#define IMAGE _w->image

static void CreateTheImage(_w)
ImageWidget _w;
{
  unsigned char *imageData, *imagePtr, *currPtr;
  int bytesPerLine;
  int i, j;

  if (!IMAGE.data)
  {
    MyXtWarning("ImageWidget: Data is empty!!!\n");
    return;
  }

  switch (_w->core.depth)
  {
    case 32:
    case 24:
    case 16: 
      MyXtWarning("ImageWidget: Unable to handle depth of %d.", _w->core.depth);
      IMAGE.ximage = (XImage *)NULL;
      return;
    case 8:
/*
** In this case since everything is already a byte in length, just create it
** using the IMAGE.data as the image data. 
*/
      imageData = (char *)XtMalloc(sizeof(char)*IMAGE.width*IMAGE.height);
      bcopy(IMAGE.data, imageData, IMAGE.width * IMAGE.height); 
      IMAGE.ximage = XCreateImage(XtDisplay(_w), IMAGE.visual, _w->core.depth, ZPixmap, 0,
        (char *)imageData, IMAGE.width, IMAGE.height, 8, 0);
      break;

    case 6:
/*
** Need information from the image, (the bytes_per_line), so just create
** the image first giving it no data to use.
*/
      IMAGE.ximage = XCreateImage(XtDisplay(_w), IMAGE.visual, _w->core.depth, ZPixmap,
        (char *)NULL, IMAGE.width, IMAGE.height, 8, 0);

      if (IMAGE.ximage->bits_per_pixel != 8)
      {
        MyXtWarning("ImageWidget: Display is %d deep, but I can't handle %d bits per pixel!",
          _w->core.depth, IMAGE.ximage->bits_per_pixel);
        XDestroyImage(IMAGE.ximage);
      }
/*
** Now just put the data so the XImage can read it.
*/
      bytesPerLine = IMAGE.ximage->bytes_per_line;
      imageData = (unsigned char *)XtCalloc(bytesPerLine * IMAGE.height, sizeof(unsigned char));
      if (!imageData)
      {
        MyXtWarning("ImageWidget: (Internal) Unable to create XImage imageData.");
        XDestroyImage(IMAGE.ximage);
        return;
      }
      IMAGE.ximage->data = (char *)imageData;
      for(i=0,imagePtr=imageData,currPtr=IMAGE.data;i<IMAGE.width*IMAGE.height;i++,imagePtr++)
        *imagePtr = *currPtr++;
    break;

    case 4:
    {

      unsigned char *bytePtr;
printf("4 bit\n");
/*
** Need information from the image, (the bytes_per_line), so just create
** the image first giving it no data to use.
*/
      IMAGE.ximage = XCreateImage(XtDisplay(_w), IMAGE.visual, _w->core.depth, ZPixmap,
        NULL, IMAGE.width, IMAGE.height, 8, 0);

      bytesPerLine = IMAGE.ximage->bytes_per_line;

      if (IMAGE.ximage->bits_per_pixel == 4)
      {
/*
** Now create enough space to put the info, use Calloc so everything is 0'd
*/

        imageData = (unsigned char *)XtCalloc(bytesPerLine * IMAGE.height, sizeof(unsigned char));
        if (!imageData)
        {
          MyXtWarning("ImageWidget: (Internal) Unable to create XImage imageData.");
          XDestroyImage(IMAGE.ximage);
          return;
        }
        IMAGE.ximage->data = (char *)imageData;

/*
** Each byte will contain two pixels' info.  Looks confusing, but
** it's not.  Just loop through the data and shift the bytes
** appropriately and put it into the new imageData array,
** where two bytes of info are put into one byte in imageData.
*/
        for(j=0,currPtr=IMAGE.data,imagePtr=imageData;j<IMAGE.height;
          j++,imagePtr+=bytesPerLine)
          for(i=0,bytePtr=imagePtr;i<IMAGE.width;i++,currPtr++)
            if (ImageByteOrder(XtDisplay(_w)) == LSBFirst)
              if (j&1) /* for 2nd halfbyte */
              {
                *bytePtr |= (*currPtr)&0xF;
                bytePtr++;
              }
              else
                *bytePtr = ((*currPtr)>>4)&0xF;
            else
              if (j&1) /* for 1st halfbyte */
              {
                *bytePtr |= ((*currPtr)>>4)&0xF;
                bytePtr++;
              }
              else
                *bytePtr = (*currPtr)&0xF;
      }
      else
      if (IMAGE.ximage->bits_per_pixel == 8)
/*
** Just use the data.
*/
        IMAGE.ximage = XCreateImage(XtDisplay(_w), IMAGE.visual, _w->core.depth, ZPixmap, 0,
          (char *)IMAGE.data, IMAGE.width, IMAGE.height, 8, 0);
      else
        MyXtWarning("ImageWidget: Display is %d deep, but I can't handle %d bits per pixel!",
          _w->core.depth, IMAGE.ximage->bits_per_pixel);
      break;
    } 
    case 1:
      MyXtWarning("I wish John would finish me at depth 1!!!\n");
      break;
    default:
      MyXtWarning("ImageWidget: Unknown depth\n");
      break;
  } 
 
  if (!IMAGE.ximage)
    MyXtWarning("ImageWidget: Unable to create ximage.\n");

  printf("%d core depth\n", _w->core.depth);
}

static void ToLower(char *_str1, char *_str2, int _length)
{
  int i;
  char *ptr;

  for(ptr=_str1,i=0;(ptr!=NULL) && (i<_length);ptr++,i++)
    *(_str2+i) = tolower(*ptr);
}

static Boolean CvtStringToGravity(_display, _args, _numArgs, _from, _to, _data)
Display *_display;
XrmValuePtr _args;
Cardinal *_numArgs;
XrmValuePtr _from;
XrmValuePtr _to;
XtPointer *_data;
{
  char *lower;
  static int gravityType;
  Boolean badConversion = FALSE;

  if (*_numArgs != 0)
  {
    XtErrorMsg("cvtStringToGravity", "wrongParamaters",
    "ImageWidget",
    "CvtStringToGravity needs no arguments.",
    (String *)NULL, (Cardinal *)NULL);
  }

  lower = XtNewString(_from->addr);
  ToLower(_from->addr, lower, strlen(_from->addr));

  if (!strcmp(lower, "forget"))
    gravityType = ForgetGravity; 
  else
  if (!strcmp(lower, "northwest"))
    gravityType = NorthWestGravity; 
  else
  if (!strcmp(lower, "north"))
    gravityType = NorthGravity; 
  else
  if (!strcmp(lower, "northeast"))
    gravityType = NorthEastGravity; 
  else
  if (!strcmp(lower, "east"))
    gravityType = EastGravity; 
  else
  if (!strcmp(lower, "southeast"))
    gravityType = SouthEastGravity; 
  else
  if (!strcmp(lower, "south"))
    gravityType = SouthGravity; 
  else
  if (!strcmp(lower, "southwest"))
    gravityType = SouthWestGravity; 
  else
  if (!strcmp(lower, "west"))
    gravityType = WestGravity; 
  else
  if (!strcmp(lower, "center"))
    gravityType = CenterGravity;
  else
  if (!strcmp(lower, "static"))
    gravityType = StaticGravity; 
  else
    badConversion = TRUE;

  XtFree(lower);

  if (badConversion)
    XtDisplayStringConversionWarning(_display, _from->addr, XtRGravity);
  else
  {
    if (_to->addr == NULL)
      _to->addr = (caddr_t)&gravityType;
    else
    if (_to->size < sizeof(int))
      badConversion = TRUE;
    else
      *(int *)_to->addr = gravityType;
      _to->size = sizeof(int);
  }

  return !badConversion;
}

/* External routines */

/* This is the reverse-engineer of CreateTheImage */

#if 0
XtPointer ImageGetData(Widget _w)
{
  ImageWidget iw;

  if (!IsImage(_w))
  {
    printf("ImageWidget: ImageGetData sent non-image widget!");
    return (XtPointer)NULL;
  }

  iw = (ImageWidget)_w;


  switch(_w->core.depth)
  {
    char *data, *imagePtr, *dataPtr;
    int i;

    case 32:
    case 24:
    case 16:
      return (XtPointer)NULL;
    case 8:

/* Just return the data. */

      data = (char *)XtMalloc(sizeof(char)*_w->image.width * _w->image.height);
      if (!data)
      {
        MyXtWarning("ImageWidget: (Internal) Insufficient resources to get data.\n");
        return (XtPointer)NULL;
      }
      bcopy(_w->image.ximage->data, data, _w->image.width * _w->image.height);
      return (XtPointer)data;
    case 6:
      data = (char *)XtMalloc(sizeof(char)*_w->image.width * _w->image.height);
      if (!data)
      {
         MyXtWarning("ImageWidget: (Internal) Insufficient resources to get data.\n");
         return (XtPointer)NULL;
      }
   
      for(i=0,imagePtr=_w->image.ximage->data,dataPtr=data;
          i<_w->image.width*_w->image.height;
          imagePtr++,i++,dataPtr++)
        *dataPtr = *imagePtr; 

      return (XtPointer)data;
    case 4:

    case 1:
  }
}

#endif

